%% generateRulePredicates(+SYMRULES,-INSTRULES)
%
%  Instantiate the list of symbolic rules. Assumes that predicate isActive/1 holds 
%  for all active (both reachable and generating) nonterminals
%
%  @param SYMRULES list of symbolic rules
%  @param INSTRULES list of instantiated rules
%

generateRulePredicates([],[]).
generateRulePredicates([H|T],RES):-
  generateRulePredicate(H,OUTR),
  %(debug->writeInstRuleInfo(H,OUTR);true),
  generateRulePredicates(T,RT),
  append(OUTR,RT,RES).


f1Generator(FNAME,ARG,TERM):-TERM=..[FNAME,ARG].

generateRulePredicates(GROUNDEDNTERMS,SYMBOLICRULES,SORTEDINSTANTIATEDRULES):-
   maplist(f1Generator(isActiveNT),GROUNDEDNTERMS,ISACTIVE), 
   retractall(isActiveNT(_)),installfacts(ISACTIVE),
   generateRulePredicates(SYMBOLICRULES,INSTANTIATEDRULES),
   sort(INSTANTIATEDRULES,SORTEDINSTANTIATEDRULES).

%% isActive(+SYMBOL)
%
%  Holds for all symbols of the grammar that are reachable and generating.
%  Holds for terminals and silent symbol and assumes the dynamic predicate
%  isActiveNT holding for all reachable and generating nonterminals. 
%
isActive(eps):-!.
isActive(term(_)):-!.
isActive(X):-isActiveNT(X).


%% generateRulePredicate(+SYMRULE,-INSTRULES,-USEDNTERMS)
%
%  Instantiate the given symbolic rule. Assumes that predicate isActive/1 holds 
%  for all active (both reachable and generating) nonterminals. The 
%
%  The instantiated rule uses only active non-terminals and does not a body.
%
%  @param SYMRULE symbolic rule
%  @param INSTRULES list of instantiated rules induced by the input rule.
%
generateRulePredicate((rule(NT,P,Q):-B),R):-!,
  findall(rule(NT,P,Q),(abstRule(NT,P,Q),isActive(NT),isActive(P),isActive(Q),B),R).

generateRulePredicate((rule(NT,A):-B),R):-!,
  findall(rule(NT,A),(abstRule(NT,A),isActive(NT),isActive(A),B),R).

generateRulePredicate(RULE,RULES):-generateRulePredicate((RULE:-true),RULES).
%generateRulePredicate(A,B):-writeln(('TRACE:',A,B)),trace.

getRHSNonTerminals(RULES,NTERMS):-
  getRHSNonTerminals1(RULES,NTERMS1),
  sort(NTERMS1,NTERMS).

getRHSNonTerminals1([],[]).
getRHSNonTerminals1([H|T],RES):-
  getRHSNonTerminals1(H,R1),
  getRHSNonTerminals1(T,R2),
  append(R1,R2,RES).

getRHSNonTerminals1(rule(_,A),[A]):-isNonTerminal(A),!.
getRHSNonTerminals1(rule(_,_),[]).
getRHSNonTerminals1(rule(_,A,B),[A,B]):-isNonTerminal(A),isNonTerminal(B),!.
getRHSNonTerminals1(rule(_,A,_),[A]):-isNonTerminal(A),!.
getRHSNonTerminals1(rule(_,_,B),[B]):-isNonTerminal(B),!.
getRHSNonTerminals1(rule(_,_,_),[]).




modifyMemInNT(MODIFYPREDICATE,init(init(PC,MEMIN),P0),init(init(PC,MEMOUT),P0)):-modifyMem(MODIFYPREDICATE,MEMIN,MEMOUT).
modifyMemInNT(MODIFYPREDICATE,pr(init(PC,MEMIN),P0,P1),pr(init(PC,MEMOUT),P0,P1)):-modifyMem(MODIFYPREDICATE,MEMIN,MEMOUT).
modifyMemInNT(MODIFYPREDICATE,init(pr(PC,MEMIN1,MEMIN2),P0),init(pr(PC,MEMOUT1,MEMOUT2),P0)):-
    modifyMem(MODIFYPREDICATE,MEMIN1,MEMOUT1),modifyMem(MODIFYPREDICATE,MEMIN2,MEMOUT2).

modifyMemInNT(MODIFYPREDICATE,pr(pr(PC,MEMIN1,MEMIN2),P0,P1),pr(pr(PC,MEMOUT1,MEMOUT2),P0,P1)):-
    modifyMem(MODIFYPREDICATE,MEMIN1,MEMOUT1),modifyMem(MODIFYPREDICATE,MEMIN2,MEMOUT2).

modifyMemInNT(_,eps,eps).
modifyMemInNT(_,term(X),term(X)).

modifyMem(MODIFYPREDICATE,MEMIN,MEMOUT):-call(MODIFYPREDICATE,MEMIN,MEMOUT).

replaceFirstTopByVar(errmem,errmem):-!.
replaceFirstTopByVar(s(GLOBIN,LOC),s(GLOBOUT,LOC)):-replaceFirstTopByVar1(GLOBIN,GLOBOUT).
replaceFirstTopByVar(s(GLOB,LOCIN),s(GLOB,LOCOUT)):-replaceFirstTopByVar1(LOCIN,LOCOUT).

replaceFirstTopByVar1(IN,OUT):-IN=..[F|ARGS],replaceFirstTopByVar2(ARGS,ARGSOUT),OUT=..[F|ARGSOUT].
replaceFirstTopByVar2([top|T],[_|T]):-!.
replaceFirstTopByVar2([H|T1],[H|T2]):-H\=top,replaceFirstTopByVar2(T1,T2),!.


replaceRuleName(NEWNAME,RULETERM,NEWRULETERM):-
    RULETERM=..[_|ARGS],
    maplist(modifyMemInNT(replaceFirstTopByVar),ARGS,ARGS1),
    NEWRULETERM=..[NEWNAME|ARGS1],!.

replaceRuleName(_,RULETERM,_):-writeln(RULETERM),trace.

generateAbstractRulePredicates(GROUNDEDNTERMS,SYMBOLICRULES,NINSTANTIATEDRULES):-
    generateRulePredicates(GROUNDEDNTERMS,SYMBOLICRULES,INSTANTIATEDRULES),
    maplist(replaceRuleName(abstRule),INSTANTIATEDRULES,NINSTANTIATEDRULES).
    


