
test(TDR):-
  createTaskDefinitionRulesFromBP('C:\\projects\\booleanProgs\\pierre\\001\\satabs.1.bp',TDR).


parseBPFile(RFILENAME,OUT):-
  ((clause(threads(_),BODY,CLREF),BODY\=fail)->
    clause_property(CLREF,file(BPFN)),
    file_directory_name(BPFN,BPDIR),
    concat_atom([BPDIR,'/',RFILENAME],FILENAME)
    ;
    FILENAME=RFILENAME
  ),
  writeln('Parsing':FILENAME),
  open(FILENAME, read, FD),
  read_stream_to_codes(FD,CODES),
  phrase(tokenize(TOKENS),CODES,R1),!,
  (R1=[T1|_] -> char_code(AT1,T1),write('unknown token':AT1),fail;true),
  phrase(program(OUT),TOKENS,R2),!,
  (R2=[_|_] -> write('Remaining tokens':R2),fail;true)
  .

createTaskDefinitionRulesFromBP(FILENAME,TDR):-
  parseBPFile(FILENAME,BP),writeln('  ... file parsed'),
  getCallGraph(BP,CG),
  getInitBP(BP,CG,INITBP),
  getThreadBPs(BP,CG,TBPS),
  (INITBP\=none->
    createInitThreadDefinition(INITBP,IT),
    TDR=[IT|TDEFS1]
  ;
    TDR=TDEFS1
  ),
  createThreadDefinitions(TBPS,TDEFS1),
  length(TDR,TDRC),writeln('no of threads':TDRC),!
  %,writeln(TDR)
  .



createThreadDefinitions(BP,T):-
  createThreadDefinitions(0,BP,T).

createThreadDefinitions(_,[],[]).
createThreadDefinitions(IDX,[BP|BPT],[T|TT]):-
  succ(IDX,IDX2),
  concat_atom(['thread',IDX],ENTRY),
  createThreadDefinition(inactive,IDX,ENTRY,BP,T),
  createThreadDefinitions(IDX2,BPT,TT).

createInitThreadDefinition(BP,T):-
  createThreadDefinition(inactive,init,init,BP,T).

createThreadDefinition(INACTIVE,TID,ENTRY,bp(GDEF,FCNLIST),THREAD):-
  assignLabels(bp([initDone|GDEF],FCNLIST),LP,LMAP),
  bp2memdef(LP,memDef(GD,LD)),
  bp2rules(LP,RULES,LMAP),
  concat_atom([ENTRY,'_',0],INITPC),
  (INACTIVE=inactive->INIT=z(INITPC);INIT=INITPC),
  getTerminalsAndNonterminals(RULES,T,_),
  checkTerminals(T,CROSSOPS),
  THREAD=thread(TID,cfGrm(INIT,RULES),memDef(GD,LD,CROSSOPS)),
  writeln(TID:INIT:memDef(GD,LD,CROSSOPS)).

supportedOperation(true).
supportedOperation(false).
supportedOperation(assert(false)).
supportedOperation(lock(system)).
supportedOperation(unlock(system)).
supportedOperation(_=true).
supportedOperation(_=false).
supportedOperation(_==true).
supportedOperation(_==false).
supportedOperation(and(X,Y)):-supportedOperation(X),supportedOperation(Y).
supportedOperation(or(X,Y)):-supportedOperation(X),supportedOperation(Y).


crossOp(A=B,[A=B],[]):-!,atom(A),atom(B).
crossOp(A==B,[A==B],[]):-!,atom(A),atom(B).
crossOp(and(A,B),CROSS,UNSUPP):-!,crossOp(A,AC,AU),crossOp(B,BC,BU),append(AC,BC,CROSS),append(AU,BU,UNSUPP).
crossOp(or(A,B),CROSS,UNSUPP):-!,crossOp(A,AC,AU),crossOp(B,BC,BU),append(AC,BC,CROSS),append(AU,BU,UNSUPP).
crossOp(X,[],X).


checkTerminals(T,CROSSOPS):-
  maplist(arg(1),T,OPS),
  exclude(supportedOperation,OPS,T2),
  maplist(crossOp,T2,CROSSOPSL,UNSUPPL),
  append(CROSSOPSL,CROSSOPS),
  append(UNSUPPL,UNSUP),
  (UNSUP=[]->
    true
  ;
    writeln('Unsupported operations:'),
    writeList(UNSUP)
  ).

modifyInitToNotifyOthers(bp(G,FNLIST),bp(G,FNLIST2)):-
  modifyInitToNotifyOthers(FNLIST,FNLIST2).

modifyInitToNotifyOthers([],[]).
modifyInitToNotifyOthers([fn(I,P,R,L,B)|T],[fn(I,P,R,L,B)|T2]):-I\=init,modifyInitToNotifyOthers(T,T2).
modifyInitToNotifyOthers([fn(init,P,R,L,B1)|T],[fn(init,P,R,L,B2)|T2]):-
  appendAsLastStmt(l_stmt([],passign([id(initDone)],[true])),B1,B2),
  modifyInitToNotifyOthers(T,T2).

%todo - treat (nested) returns
appendAsLastStmt(S,[],[S]).
appendAsLastStmt(S,[l_stmt(L,return(X))|T],[S,l_stmt(L,return(X))|T]):-!.
appendAsLastStmt(S,[H|T],[H|T2]):-appendAsLastStmt(S,T,T2).

getInitBP(BPALL,CG,INITBP2):-
  getFunctionClosureBP(init,CG,BPALL,INITBP)->
    modifyInitToNotifyOthers(INITBP,INITBP2)
  ;
  INITBP=none.

getThreadBPs(BPALL,CG,TBPS):-
  getThreadBPs(BPALL,CG,0,TBPS).

getThreadBPs(BPALL,CG,IDX,TBPS):-
  concat('thread',IDX,F),
  getFunctionClosureBP(F,CG,BPALL,TBP)->
    succ(IDX,IDX2),
    getThreadBPs(BPALL,CG,IDX2,TBPS2),
    TBPS=[TBP|TBPS2]
  ;
    TBPS=[]
  .


functionSelectorByNameSet(SET,fn(NAME,_,_,_,_)):-member(NAME,SET).

getFunctionClosureBP(ENTRYFCN,CG,bp(GLOB,FCNLIST),bp(GLOB,FCNLIST2)):-
  getFunctionClosure(ENTRYFCN,CG,CLOSURE),
  include(functionSelectorByNameSet(CLOSURE),FCNLIST,FCNLIST2),FCNLIST2\=[].


getFunctionClosure(NAME,CG,CLOSURE):-
  getFunctionClosure(NAME,CG,[],CLOSURE).

getFunctionClosure(NAME,CG,CLIN,CLOUT):-
  member(NAME,CLIN)->
    CLOUT=CLIN
  ;
    getTargets(NAME,CG,TARGETS),
    getFunctionClosureIt(NAME,CG,TARGETS,[NAME|CLIN],CLOUT)
  .

getFunctionClosureIt(_,_,[],CLIN,CLIN).
getFunctionClosureIt(NAME,CG,[TRG|T],CLIN,CLOUT):-
  getFunctionClosure(TRG,CG,CLIN,CLOUT1),
  getFunctionClosureIt(NAME,CG,T,CLOUT1,CLOUT).
  

targetSelector(NAME,NAME-TARGET,TARGET).
targetSelector(NAME,OTHERNAME-_,nop):-NAME\=OTHERNAME.

getTargets(NAME,CG,TARGETS):-
  maplist(targetSelector(NAME),CG,TARGETS1),
  delete(TARGETS1,nop,TARGETS).


getCallGraph(bp(_,FNLIST),CG):-
  getCallGraphFn(FNLIST,[],CGLIST),
  sort(CGLIST,CG).

getCallGraphFn([],IN,IN).
getCallGraphFn([fn(NAME,_,_,_,BODY)|T],IN,OUT):-
	getCallGraph(NAME,BODY,IN,OUT1),
	getCallGraphFn(T,OUT1,OUT).

getCallGraph(_,[],IN,IN).
getCallGraph(NAME,[l_stmt(_,S)|T],IN,OUT):-getCallGraph1Stmt(NAME,S,IN,OUT1),getCallGraph(NAME,T,OUT1,OUT).
getCallGraph1Stmt(NAME,if(_,THEN,ELSE),IN,OUT):-!,getCallGraph(NAME,THEN,IN,OUT1),getCallGraph(NAME,ELSE,OUT1,OUT).
getCallGraph1Stmt(NAME,while(_,BODY),IN,OUT) :- !,getCallGraph(NAME,BODY,IN,OUT).
getCallGraph1Stmt(NAME,call(TNAME,_),IN,[NAME-TNAME|IN]):-!.
getCallGraph1Stmt(_,_,IN,IN).

assignLabels(IN,OUT,LMAP):-empty_assoc(E),assignLabels(IN,OUT,E,LMAP).

assignLabels(bp(D,FUNLIST),bp(D,FUNLIST2),X,Y):-assignLabels(FUNLIST,FUNLIST2,X,Y).
assignLabels([fn(NAME,PARAMS,RET,LOCALDECL,BODY1)|T1],[fn(NAME,PARAMS,RET,LOCALDECL,BODY2)|T2],M,N):-assignLabels(BODY1,BODY2,ctx(0,M),ctx(_,M1),NAME),assignLabels(T1,T2,M1,N).
assignLabels([],[],X,X).

assignLabels([],[],Z,Z,_).
assignLabels([IN|INT],[OUT|OUTT],X,Y,NAME):-assignLabels(IN,OUT,X,Z,NAME),assignLabels(INT,OUTT,Z,Y,NAME).

assignLabels(l_stmt(ORIGLABS,STMT),l_stmt(LAB,STMT2),ctx(LI1,MAP1),ctx(LI,MAP),NAME):-
	 concat_atom([NAME,LI1],'_',LAB),succ(LI1,LI2),insertLabesIntoMap(ORIGLABS,LAB,MAP1,MAP2),
	 assignLabelStmt(STMT,STMT2,ctx(LI2,MAP2),ctx(LI,MAP),NAME).

assignLabelStmt(if(COND,THEN,ELSE),if(COND,THEN2,ELSE2),X,Y,NAME):-!,assignLabels(THEN,THEN2,X,X1,NAME),assignLabels(ELSE,ELSE2,X1,Y,NAME).
assignLabelStmt(while(COND,BODY),while(COND,BODY2),X,Y,NAME) :- !,assignLabels(BODY,BODY2,X,Y,NAME).
assignLabelStmt(S,S,X,X,_).

insertLabesIntoMap([],_,M,M).
insertLabesIntoMap([H|T],L,MIN,MOUT):-put_assoc(H,MIN,L,MIN2),insertLabesIntoMap(T,L,MIN2,MOUT).


bp2memdef(bp(GDECL,FUNLIST),memDef([mutex-system|GD],LD)):-
	fn2memdef(FUNLIST,LDECL),createBoolDef(GDECL,GD),createBoolDef(LDECL,LD).

fn2memdef([fn(NAME,PARAMS,RET,LOCALD,_)|T],LD):- fn2memdef(T,LD2),
  (PARAMS\=[]->write('Function '),write(NAME),write(' has arguments - not yet supported');true),
  (RET=bool->write('Function '),write(NAME),write(' has boolean return value - not yet supported');true),
  append(LOCALD,LD2,LD).

fn2memdef([],[]).

createBoolDef([],[]).
createBoolDef([id(ID)|T],[bool-ID|RT]):-createBoolDef(T,RT).
createBoolDef([ID|T],[bool-ID|RT]):-atom(ID),createBoolDef(T,RT).

%stmt2rules(STMT/L,RULES,NEXTLABEL)

bp2rules(bp(_,FUNLIST),R,CTX):-fn2rules(FUNLIST,R,CTX).
fn2rules([fn(NAME,_,_,_,BODY)|T],[rule(FLABEL,eps)|RULES],CTX):- concat_atom([NAME,'_end'],FLABEL),stmt2rules(BODY,R1,FLABEL,CTX,_),fn2rules(T,R2,CTX),append(R1,R2,RULES).
fn2rules([],[],_).

stmt2rules([],[],FIRST,_,FIRST).
stmt2rules([l_stmt(LL,S)|[]],RULES,NEXT,CTX,LL):-stmt2rules(LL,NEXT,S,RULES,CTX).
stmt2rules([l_stmt(LL,S),l_stmt(LL2,S2)|T],RULES,NEXT,CTX,LL):-stmt2rules(LL,LL2,S,R1,CTX),stmt2rules([l_stmt(LL2,S2)|T],R2,NEXT,CTX,_),append(R1,R2,RULES).

stmt2rules(INL,OUTL,if(ndet,THEN,ELSE),[rule(INL,FIRSTTHEN),rule(INL,FIRSTELSE)|RULES],CTX):-!,stmt2rules(THEN,R1,OUTL,CTX,FIRSTTHEN),stmt2rules(ELSE,R2,OUTL,CTX,FIRSTELSE), append(R1,R2,RULES).
stmt2rules(INL,OUTL,if(CONDIN,THEN,ELSE),[rule(INL,term(COND),FIRSTTHEN),rule(INL,term(NOTCOND),FIRSTELSE)|RULES],CTX):-!,
  normalizeCond(CONDIN,COND,NOTCOND),
  stmt2rules(THEN,R1,OUTL,CTX,FIRSTTHEN),stmt2rules(ELSE,R2,OUTL,CTX,FIRSTELSE), append(R1,R2,RULES).

stmt2rules(INL,OUTL,while(ndet,BODY),[rule(INL,FIRSTBODY),rule(INL,OUTL)|RULES],CTX) :- !,stmt2rules(BODY,RULES,INL,CTX,FIRSTBODY).
stmt2rules(INL,OUTL,while(CONDIN,BODY),[rule(INL,term(COND),FIRSTBODY),rule(INL,term(NOTCOND),OUTL)|RULES],CTX) :- !,
  normalizeCond(CONDIN,COND,NOTCOND),
  stmt2rules(BODY,RULES,INL,CTX,FIRSTBODY).

stmt2rules(L1,L2,passign(LHSL,RHSL),RULES,_):-!,
  passignAndLoop(LHSL,RHSL,PASSIGNEXPR),
  maplist(createPassignRule(L1,L2),PASSIGNEXPR,RULES).



stmt2rules(L1,L2,assert(E),[rule(L1,term(assert(ES)),L2)],_) :-simplify(E,ES),!.
stmt2rules(L1,L2,print(_),[rule(L1,L2)],_) :-!.
stmt2rules(L1,L2,call(NAME,[]),[rule(L1,FLABEL,L2)],_) :-concat_atom([NAME,'_0'],FLABEL),!.

stmt2rules(L1,_,return([]),[rule(L1,eps)],_) :-!.
stmt2rules(L1,L2,skip,[rule(L1,L2)],_) :-!.
stmt2rules(L1,_,goto(L),[rule(L1,TARGET)],MAP) :- get_assoc(L,MAP,TARGET),!.
stmt2rules(L1,_,goto([L]),[rule(L1,TARGET)],MAP) :- get_assoc(L,MAP,TARGET),!.
stmt2rules(L1,L2,atomic_begin,[rule(L1,term(lock(system)),L2)],_) :-!.
stmt2rules(L1,L2,atomic_end,[rule(L1,term(unlock(system)),L2)],_) :-!.
stmt2rules(L1,L2,assume(E),[rule(L1,term(ES),L2)],_) :-normalizeCond(E,ES,_),!.
stmt2rules(_,_,S,[],_):-write('Unsupported statement:'),writeln(S).

createPassignRule(L1,L2,PASSIGNEXPR,rule(L1,term(PASSIGNEXPR),L2)).

passignExpr(id(LHS),RHS,ASGNS):-simplify(RHS,RHSS),
  (RHSS=schoose(RHS1,RHS2)->
    ASGNS=[LHS=RHS1,LHS=RHS2]
  ;
    ASGNS=[LHS=RHSS]
  ).

createAndTerm(X,Y,and(X,Y)).
multiplyAnds([],_,[]).
multiplyAnds([H|T],ASGNS,M):-
  maplist(createAndTerm(H),ASGNS,N1),
  multiplyAnds(T,ASGNS,N2),
  append(N1,N2,M).

passignAndLoop([LHS],[RHS],ASGNS):-passignExpr(LHS,RHS,ASGNS).
passignAndLoop([LHS|LHST],[RHS|RHST],ASGNS):-LHST\=[],RHST\=[],passignExpr(LHS,RHS,ASGNS1),
  passignAndLoop(LHST,RHST,ASGNS2),
  multiplyAnds(ASGNS1,ASGNS2,ASGNS).

simplify(true,true):-!.
simplify(false,false):-!.
simplify(id(X),X):-!.
simplify(X,X):-atom(X),!.
simplify(id(X)==true,X==true).
simplify(id(X)==false,X==false).
simplify(X==true,X==true).
simplify(X==false,X==false).


simplify(not(and(O1,O2)),RES):-!,simplify(not(O1),O1S),simplify(not(O2),O2S),simplify1(or(O1S,O2S),RES).
simplify(not(or(O1,O2)),RES):-!,simplify(not(O1),O1S),simplify(not(O2),O2S),simplify1(and(O1S,O2S),RES).
simplify(and(O1,O2),RES):-simplify(O1,O1S),simplify(O2,O2S),simplify1(and(O1S,O2S),RES).
simplify(or(O1,O2),RES):-simplify(O1,O1S),simplify(O2,O2S),simplify1(or(O1S,O2S),RES).
simplify(not(O),RES):-simplify(O,OS),simplify1(not(OS),RES).
simplify(schoose(O1,O2),RES):-simplify(O1,O1S),simplify(O2,O2S),simplify1(schoose(O1S,O2S),RES).

simplify1(and(true,true),true):-!.
simplify1(and(false,_),false):-!.
simplify1(and(_,false),false):-!.
simplify1(and(O1,O2),and(O1,O2)).

simplify1(or(false,false),false):-!.
simplify1(or(true,_),true):-!.
simplify1(or(_,true),true):-!.
simplify1(or(O1,O2),or(O1,O2)).

simplify1(not(true),false):-!.
simplify1(not(false),true):-!.
simplify1(not(not(E)),E):-!.
simplify1(not(X==true),X==false):-!.
simplify1(not(X==false),X==true):-!.
simplify1(not(E),not(E)).

simplify1(schoose(A,B),A):-A==B,!.
simplify1(schoose(A,B),schoose(A,B)).

normalizeCond(X,COND,NOTCOND):-
  simplify(X,Y),normalizeCond_hlp(Y,COND,NOTCOND).

normalizeCond_hlp(true,true,false):-!.
normalizeCond_hlp(false,false,true):-!.
normalizeCond_hlp(not(X),COND,NOTCOND):-!,normalizeCond(X,NOTCOND,COND).
normalizeCond_hlp(X==true,X==true,X==false):-!.
normalizeCond_hlp(X==false,X==false,X==true):-!.

normalizeCond_hlp(Y,COND,NOTCOND):-atom(Y),!,COND=(Y==true),NOTCOND=(Y==false).
normalizeCond_hlp(Y,Y,NOT_Y):-simplify(not(Y),NOT_Y).



%unsupported:
%
%stmt2rules(sync(VAR)) :-!,
%
%
%stmt2rules(L1,L2,assume(E),[],_) :-!,
%stmt2rules(dead(IDL)) :-!,
%stmt2rules(start_thread) :-!,
%stmt2rules(end_thread) :-!,
%stmt2rules(L1,L2,call(NAME,PARS),[],_) :-!,
%stmt2rules(L1,L2,return(L),[],_) :-!,

%**GRAMMAR**********

program(bp(DECLLIST,FUNLIST)) --> pstart, g_decl_list(DECLLIST) , fun_list(FUNLIST), pend.

pstart --> {true}.

pend --> {true}.

%/***** decls *****/

g_decl_list(DECLLIST) --> ['decl'], g_identifier_list(IDLIST), [';'] ,!, g_decl_list(DECLLIST2),{append(IDLIST,DECLLIST2,DECLLIST)}.
g_decl_list([]) --> {true}.

g_identifier_list([H|T]) --> g_decl_id(H), [','], g_identifier_list(T).
g_identifier_list([ID]) --> g_decl_id(ID).

g_decl_id(id(ID)) --> [id(ID)].
	
decl_list(DECLLIST) --> ['decl'], identifier_list(IDLIST),[';'], !, decl_list(DECLLIST2),{append(IDLIST,DECLLIST2,DECLLIST)}.
decl_list([]) --> {true}.

identifier_list([H|T]) --> decl_id(H), [','], identifier_list(T).
identifier_list([ID]) --> decl_id(ID).

decl_id(id(ID))	--> [id(ID)].
	  
%/***** functions *****/

fun_list([FD|FDL]) --> function_definition(FD), fun_list(FDL).
fun_list([FD]) --> function_definition(FD).


function_definition(fn(NAME,PARAMS,RET,LOCALDECL,BODY)) --> function_heading(NAME,RET), parms(PARAMS) , ['begin'] ,decl_list(LOCALDECL), enforce, abortif, statement_list(BODY), ['end'].
function_heading(NAME,RET) --> fun_type(RET), [id(NAME)].


fun_type(bool) -->  fun_search_order, ['bool'].
fun_type(bool) -->  fun_search_order, ['bool'], ['<'], [num(_)], ['>'].
fun_type(void) -->  fun_search_order, ['void'].

fun_search_order --> ['dfs'].
fun_search_order --> {true}.

enforce(E) --> ['enforce'], expression(E), [';'].
enforce --> {true}.

abortif(E) --> ['abortif'], expression(E),[';'].
abortif --> {true}.

parms(DECLLIST) --> ['('], parameter_list(DECLLIST), [')'].

parameter_list(DECLLIST) -->  identifier_list(DECLLIST).
parameter_list([]) --> {true}.

%/***** statements *****/

statement_list([S|T]) --> labelled_statement(S), !, statement_list(T).
statement_list([]) --> [';'];{true}.

labelled_statement(l_stmt(LL,C)) --> label_list(LL), concurrent_statement(C).
labelled_statement(l_stmt([],C)) --> concurrent_statement(C).

label_list([l(L)]) --> [id(L),':'].
label_list([l(L)|LL]) --> [id(L),':'],!, label_list(LL).


target_label_list([l(L)]) --> [id(L)].
target_label_list([l(L)]|LL) --> [id(L),','],target_label_list(LL).


concurrent_statement(S) --> statement(S).
concurrent_statement(S) --> start_thread_statement(S),[';'].
concurrent_statement(S) --> end_thread_statement(S),[';'].
concurrent_statement(S) --> sync_statement(S),[';'].
concurrent_statement(S) --> atomic_begin_statement(S),[';'].
concurrent_statement(S) --> atomic_end_statement(S),[';'].
        

statement(S) --> dead_statement(S),[';'].
statement(S) --> parallel_assign(S),[';'].
statement(S) --> assert_nt(S),[';'].
statement(S) --> assume(S),[';'].
statement(S) --> print_nt(S),[';'].
statement(S) --> call_nt(S),[';'].
statement(S) --> selection_statement(S).
statement(S) --> iteration_statement(S).
statement(S) --> jump_statement(S),[';'].

dead_statement(dead(IDL)) --> ['dead'], assign_id_list(IDL).
	   
parallel_assign(passign(LHSL,RHSL)) --> assign_id_list(LHSL), [':='],(assign_expression_list(RHSL), assign_constrain_opt
                                                                      ;call_nt(RHS),{RHSL=[RHS],writeln('Function calls on RHS not supported!':RHS)}). 
    
assign_constrain_opt --> {true}.
assign_constrain_opt(E) --> ['constrain'], expression(E).

assign_expression_list([H|T]) --> choose_expression(H),([','],!,assign_expression_list(T);{T=[]}).
assign_id_list([ID]) --> asgn_id(ID).
assign_id_list([ID|T]) --> asgn_id(ID), [','], assign_id_list(T)  .
asgn_id(id(ID))--> [id(ID)].
assert_nt(assert(E)) --> ['assert'], expression(E).
assume(assume(E)) --> ['assume'], expression(E).
print_nt(print(EL)) --> ['print'],['('], expression_list(EL), [')'].
call_nt(call(NAME,PARS)) --> [id(NAME)], ['('], expression_list(PARS), [')'].
call_nt(call(NAME,[])) --> [id(NAME),'(',')'].
selection_statement(if(COND,THEN,ELSE)) --> ['if'], choose_expression(COND), ['then'], statement_list(THEN), elsif_list(ELSE).
elsif_list([]) -->  ['fi'],([';'];{true}).
elsif_list(SL) -->  ['else'], statement_list(SL), ['fi'],([';'];{true}).
elsif_list([l_stmt([],if(COND,THEN,ELSE))]) -->  ['elsif'], choose_expression(COND), ['then'] , statement_list(THEN) ,elsif_list(ELSE).
iteration_statement(while(COND,BODY)) --> ['while'], expression(COND), ['do'], statement_list(BODY), ['od'],([';'];{true}).
jump_statement(return([])) --> ['return'].
jump_statement(return(L)) --> ['return'], expression_list(L).
jump_statement(skip) --> ['skip'].
jump_statement(goto(L)) --> ['goto'], target_label_list(L).

start_thread_statement(start_thread) --> ['start_thread'], ['goto'], [id(_)].
start_thread_statement(start_thread) --> ['start_thread'], [id(_),'(',')'].

end_thread_statement(end_thread) --> ['end_thread'].

sync_statement(sync(VAR)) --> ['sync',id(VAR)].

atomic_begin_statement(atomic_begin) --> ['atomic_begin'].

atomic_end_statement(atomic_end) --> ['atomic_end'].


%/***** expressions *****/

primary_expression(E) --> ['('], expression(E), [')'].
primary_expression(ndet) --> ['*'].
primary_expression(schoose(L,R)) --> ['schoose'], ['['], expression(L) ,[','], expression(R), [']'].
primary_expression(id(V)) --> [id(V)].
primary_expression(primed(V)) --> ['\'', id(V)].
primary_expression(N) --> [num(N1)],{N1==0->N=false;N=true}.
primary_expression(true) --> ['T'].
primary_expression(false) --> ['F'].
	
unary_expression(neg(E)) --> ['~'],!,primary_expression(E).
unary_expression(not(E)) --> ['!'],!,primary_expression(E).
unary_expression(E) --> primary_expression(E).

equality_expression(E) --> unary_expression(E1),
      (['='],!, equality_expression(E2),{E=(E1==E2)}
      ;['!='],!,equality_expression(E3),{E=not(E1==E3)}
      ;{E=E1}).

true-->{true}.

and_expression(E) --> equality_expression(E1), (['&&'],!, and_expression(E2),{E=and(E1,E2)};{E=E1}).

exclusive_or_expression(E) --> and_expression(E1), (['^'],!, exclusive_or_expression(E2),{E=xor(E1,E2)};{E=E1}).


or_expression(E) --> exclusive_or_expression(E1), 
                     (['||'],!, or_expression(E2),{E=or(E1,E2)}
                     ;['=>'],!, or_expression(E3),{E=impl(E1,E3)}
                     ;{E=E1}).
	
expression(E) --> or_expression(E1), 
               (['?'], expression(E2), [':'], or_expression(E3), {E=cexp(E1,E2,E3)} 
               ;{E=E1}).

choose_expression(E) --> expression(E).

expression_list(EL) --> choose_expression(E), ([','],!, expression_list(E2),{EL=[E|E2]};{EL=[E]}).

%#*********Tokens**********


tokenize([T|L])-->token(T),!,tokenize(L).
tokenize([T|L])-->token_integer(T),!,tokenize(L).
tokenize([NT|L])-->token_identifier(T),{ttoken(T)->NT=T;NT=id(T)},!,tokenize(L).
tokenize(L) --> whiteSpace,!,whiteSpaces,tokenize(L).
tokenize(L) --> comment,!,tokenize(L).
tokenize([]) --> {true}.

whiteSpace-->[C],{code_type(C,space)}.
whiteSpaces-->whiteSpace,!,whiteSpaces.
whiteSpaces-->{true}.

comment --> "//",whateverTillNl.
whateverTillNl --> [NL],{code_type(NL,end_of_file),!}.
whateverTillNl --> [NL],{code_type(NL,end_of_line),!}.
whateverTillNl --> [_],whateverTillNl.

digit(D) --> [D], { code_type(D, digit) }.
digits([D|T]) --> digit(D), !,digits(T).
digits([]) --> [].
token_integer(num(I)) --> digit(D0),digits(D), { number_chars(I, [D0|D]) }.

letter(L) --> [L], { code_type(L,alnum) }.
letter(C) --> "_", {atom_char('_',C)}.
letters([H|T])--> letter(H),!,letters(T).
letters([])-->[].

token_identifier(A) --> letter(L0),letters(L), {atom_chars(A,[L0|L])}.

ttoken('return').
ttoken('goto').
ttoken('skip').
ttoken('dead').
ttoken('dfs').
ttoken('do').
ttoken('od').
ttoken('else').
ttoken('elsif').
ttoken('if').
ttoken('then').
ttoken('fi').
ttoken('decl').
ttoken('while').
ttoken('begin').
ttoken('end').
ttoken('bool').
ttoken('void').
ttoken('assert').
ttoken('assume').
ttoken('print').
ttoken('enforce').
ttoken('constrain').
ttoken('schoose').
ttoken('abortif').
ttoken('start_thread').
ttoken('end_thread').
ttoken('sync').
ttoken('atomic_begin').
ttoken('atomic_end').
ttoken('T').
ttoken('F').

token('*')-->"*".
token('?')-->"?".

token(':=')-->":=".
token('=')-->"=".
token('!=')-->"!=".
token('=>')-->"=>".
token('&&')-->"&&".
token('&&')-->"&".
token('||')-->"||".
token('||')-->"|".
token('(')-->"(".
token(')')-->")".
token(',')-->",".
token(';')-->";".
token('[')-->"[".
token(']')-->"]".
token('!')-->"!".
token(':')-->":".

