/** <pattern> Generate the regular grammar representing the given pattern 
@author Tomas Poch
*/

%:-module(pattern, [patternToRules/3]).

%% patternToRules(+PTRN,-REGGRM,-WORDCOUNT)
% 
%  For the pattern (list of lists of terms) create the regular grammar generating the same langua
%
%  @param PTRN
%  @param 
% 
patternToRules(PTRN,regGrm(p(0,0),RULES,ptrn,ENCODEINFO),WORDCOUNT):-
  patternToRules_hlp(PTRN,SYMRULES,0,WORDCOUNT,SYMBOLCOUNTS),
  generateValidRules(0,SYMBOLCOUNTS,VALIDRULES),
  generatePtrnEncodeInfo(SYMBOLCOUNTS,ENCODEINFO),
  append(SYMRULES,VALIDRULES,RULES),
  (verbose->sumSymbolCounts(SYMBOLCOUNTS,SUM),writeln('Pattern states':SUM),nl,nl;true).
  

sumSymbolCounts([],0).
sumSymbolCounts([OP|OPS],RES):-
  sumSymbolCounts(OPS,RES1),
  RES is RES1 + OP+1.

patternToRules_hlp([],[],C,C,[]).
patternToRules_hlp([W|WT],R,THISLINE,COUNT,[THISLINECOUNT|NEXTCOUNT]):-
    wordToRulesFirstItem(W,FIRSTITEMRULES,THISLINE,0),
    wordToRules(W,WORDRULES,THISLINE,0,THISLINECOUNT),
    append(FIRSTITEMRULES,WORDRULES,TMP),
    NXTLINE is THISLINE + 1,
    patternToRules_hlp(WT,WTR,NXTLINE,COUNT,NEXTCOUNT),
    append(TMP,WTR,R).


abstractTerms(go(X),[go(X)]).
abstractTerms(yield(GM),ABSTS):-
  abstractTermsYield(0,GM,ABSTS).

abstractTermsYield(IDX,GM,RES):-
  getAbstractPart(IDX,GM,ABST)->
    succ(IDX,IDX2),
    abstractTermsYield(IDX2,GM,ABSTS),
    RES=[yield(ABST)|ABSTS]
    ;
    RES=[]
  .    


termrule(IN,OUT,T,rule(IN,term(T),OUT)).

abstractAlternativeRules(IN,OUT,T,RULES):-
  (useAbstraction(all)->
    abstractTerms(T,ATS),
    maplist(termrule(IN,OUT),ATS,RULES)
    ;
    RULES=[rule(IN,term(T),OUT)]
  ).
  

wordToRulesFirstItem(_,[rule(p(M,0),eps)],M,M).
wordToRulesFirstItem([T|TT],RULES,THISLINE,M):-
    TT\=[],
    THISLINE>M,
    M1 is M + 1,
    abstractAlternativeRules(p(M,0),p(THISLINE,1),T,R1),
    wordToRulesFirstItem([T|TT],RR,THISLINE,M1),
    append(R1,RR,RULES).

wordToRulesFirstItem([SINGLE],RULES,THISLINE,M):-
    THISLINE>M,
    M1 is M + 1,
    abstractAlternativeRules(p(M,0),p(THISLINE,0),SINGLE,R1),
    wordToRulesFirstItem([SINGLE],RR,THISLINE,M1),
    append(R1,RR,RULES).

wordToRules([LAST],R1,THISLINE,M,M):-
    abstractAlternativeRules(p(THISLINE,M),p(THISLINE,0),LAST,R1).
    
wordToRules([T|TT],RULES,THISLINE,M,COUNT):-
    M1 is M + 1,
    abstractAlternativeRules(p(THISLINE,M),p(THISLINE,M1),T,R1),
    wordToRules(TT,RT,THISLINE,M1,COUNT),
    append(R1,RT,RULES).


wordToRules([],[],_,_,_).

generateValidRules(_,[],[]).
generateValidRules(IDX,[SYMCOUNT|T],RULES):-
    IDX2 is IDX+1,
    generateValidRules(IDX2,T,RULEST),
    generateValidForLine(IDX,SYMCOUNT,LINERULES),
    append(RULEST,LINERULES,RULES).

generateValidForLine(LIDX,0,[valid(ptrn,p(LIDX,0))]).
generateValidForLine(LIDX,SIDX,[valid(ptrn,p(LIDX,SIDX))|RT]):-SIDX>0,
    SIDX2 is SIDX-1,
    generateValidForLine(LIDX,SIDX2,RT).
    
generatePtrnEncodeInfo(_,MAX,[],[],MAX).
generatePtrnEncodeInfo(IDX,SUM,[SYMCOUNT|T],[ENCODE|RT],MAX):-    
    SUM2 is SUM + SYMCOUNT+1,
    ENCODE=(encodeNT(ptrn,p(IDX,X),CODE):-(CODE is SUM + X)),
    succ(IDX,IDX2),
    generatePtrnEncodeInfo(IDX2,SUM2,T,RT,MAX).

generatePtrnEncodeInfo(SYMCOUNTS,encodeInfo(RULES,MAX)):-
    generatePtrnEncodeInfo(0,0,SYMCOUNTS,RULES,MAX).


