/** <sortClauseBodies> Sort clause bodies so that valid rules goes first.
@author Tomas Poch
*/

%:-module(sortClauseBodies, [sortClauseBodies/2]).

%%  sortClauseBodies(+RULES,-RULESWITHSORTETBODIES).
%
%  Sort clause bodies so that valid rules goes first.
sortClauseBodies([],[]).
sortClauseBodies([H|T],[H2|T2]):-
    sortClauseBodies(T,T2),
    sortClauseBodies1(H,SORTED),
    (SORTED=(H1:-true)->H2=H1;H2=SORTED).

sortClauseBodies1(X,X):-X\=(_:-_).
sortClauseBodies1((H:-BODYIN),(H:-BODYOUT)):-sortBody(BODYIN,BODYOUT).

sortBody(IN,OUT):-
    toKeyList(IN,INLIST),
    keysort(INLIST,OUTLIST),
    fromKeyList(OUTLIST,OUT).


toKeyList(IN,[KEY-IN]):-IN\=(_,_),IN\=true,getKey(IN,KEY).
toKeyList(true,[]).
toKeyList((IN1,IN2),[KEY-IN1|T]):-
    IN1\=true,
    getKey(IN1,KEY),
    toKeyList(IN2,T).

toKeyList((true,IN2),T):-toKeyList(IN2,T).

getKey(T,10):-functor(T,rule,_),!.
getKey(T,20):-functor(T,valid,_),!.
getKey(T,20):-functor(T,gvalid,_),!.
getKey(T,20):-functor(T,lvalid,_),!.
getKey(_,30).

fromKeyList([],true).
fromKeyList([_-TAIL],TAIL).
fromKeyList([_-H|T],(H,REST)):-T\=[],fromKeyList(T,REST).

