/** <toolstate> Keep the tool state as a predicate in the program database

The tool modifies the state of the prolog database for two purposes
- to store the options - verbose input, turning the abstraction phase on
- to store a grammar in from of rule() predicates and traverse it
    
@author Tomas Poch
*/

%:-module(toolstate, [debug/0,
%                     groundRules/0,setRuleGrounding/0,unsetRuleGrounding/0,
%                     useAbstraction/0,setUseAbstraction/0,unsetUseAbstraction/0,
%                     printTime/0,setPrintTime/0,unsetPrintTime/0,
%                     printQueueSize/0,setPrintQueueSize/0,unsetPrintQueueSize/0,
%                     verbose/0,setVerbose/0,unsetVerbose/0,
%                     useAbstraction/0,setUseAbstraction/0,unsetUseAbstraction/0,
%		     installGrammar/1,notFailingBagOf/3,notFailingSetOf/3,
%		     writeList/1
%                    ]).

debug:-false.
%debug:-true.

setState(STATE):-retractall(STATE),assertz(STATE:-true).
unsetState(STATE):-retractall(STATE),assertz(STATE:-false).

setCurrentAbstInfo(ABSTINFO):-
  retractall(currentAbstInfo(_)),
  assertz(currentAbstInfo(ABSTINFO)).


:- dynamic(groundRules/0).
setRuleGrounding:-setState(groundRules).
unsetRuleGrounding:-unsetState(groundRules).

:- dynamic(useAbstraction/0).
setUseAbstraction(MODE):-assertz(useAbstraction(MODE)).

:- dynamic(printTime/0).
setPrintTime:-setState(printTime).
unsetPrintTime:-unsetState(printTime).

:- dynamic(printQueueSize/0).
setPrintQueueSize:-setState(printQueueSize).
unsetPrintQueueSize:-unsetState(printQueueSize).

:- dynamic(verbose/0).
setVerbose:-setState(verbose).
unsetVerbose:-unsetState(verbose).

:- dynamic(execute/0).
setExecute:-setState(execute).
unsetExecute:-unsetState(execute).


verboseWriteLn(WRITEARG):-
  verbose->writeln(WRITEARG);true.

:- dynamic(solverMethod/1).
setSolverMethod(METHOD):-retractall(solverMethod(_)),assertz(solverMethod(METHOD)).

:-unsetPrintTime.
:-unsetPrintQueueSize.
:-unsetVerbose.
:-unsetExecute.

:-setSolverMethod(yices).

installGrammar(cfGrm(_,RULES,SUPPORTRULES)):-
   append(SUPPORTRULES,RULES,ALLRULES),
   installGrammarFacts(ALLRULES).
 
installGrammarFacts(GRAMMAR):-
    retractall(rule(_,_)),
    retractall(rule(_,_,_)),
    retractall(rule(_,_,_,_)),
    retractall(valid(_)),
    retractall(valid(_,_)),
    retractall(validpc(_,_)),
    retractall(gvalid(_)),
    retractall(lvalid(_)),
    retractall(irule(_,_,_,_)),
    retractall(irule(_,_,_)),
    retractall(encodeNT(_,_)),
    retractall(encodeNT(_,_,_)),
    retractall(validRuleMap(_,_)),
    installfacts(GRAMMAR).


installfacts([]).
installfacts([H|T]):-assertz(H),installfacts(T).


:-meta_predicate notFailingBagOf(-, 0, -).
notFailingBagOf(T,CT,BAG):-
  bagof(T,CT,BAG)->true;BAG=[].

:-meta_predicate notFailingSetOf(-, 0, -).
notFailingSetOf(T,CT,BAG):-
  setof(T,CT,BAG)->true;BAG=[].


:- dynamic(rule/3).
:- dynamic(rule/2).
:- dynamic(isActiveNT/1).
:- dynamic(valid/1).
:- dynamic(valid/2).
:- dynamic(validpc/2).
:- dynamic(gvalid/1).
:- dynamic(lvalid/1).
:- dynamic(irule/4).
:- dynamic(irule/3).
:- dynamic(encodeNT/2).
:- dynamic(encodeNT/3).

%%      writeList(?List:list) is det.
%
%       Write the given list to the standard output, item by line 
%
writeList([]).
writeList([H|T]):-write(H),nl,writeList(T).
